
import os
from pathlib import Path
import glob
import warnings
from concurrent.futures import ThreadPoolExecutor

HERE = Path(__file__).parent


_COMPILER_DIR = Path("/js/compiler")


def task(src: str | Path) -> None:
    src = Path(src)
    if "entrypoint.sh" in str(src):
        return
    link_dst = Path("/js") / src.name
    
    # Handle shell scripts
    if src.suffix == '.sh':
        os.system(f"dos2unix {src} && chmod +x {src}")

    # if link exists, remove it
    if link_dst.exists():
        print(f"Removing existing link {link_dst}")
        try:
            os.remove(link_dst)
        except Exception as e:
            warnings.warn(f"Failed to remove {link_dst}: {e}")

        
    if not link_dst.exists():
        print(f"Linking {src} to {link_dst}")
        try:
            os.symlink(str(src), str(link_dst))
        except FileExistsError:
            print(f"Target {link_dst} already exists")
    else:
        print(f"Target {link_dst} already exists")


def make_links() -> None:
    # Define file patterns to include
    patterns = ['*.h', '*.py', '*.css', '*.sh', "*.ino", "*.hpp", "*.cpp", "*.ini", "*.txt"]
    
    # Get all matching files in compiler directory
    files = []
    for pattern in patterns:
        files.extend(glob.glob(str(_COMPILER_DIR / pattern)))

    # Process files in parallel using ThreadPoolExecutor
    with ThreadPoolExecutor(max_workers=16) as executor:
        executor.map(task, files)


def init_runtime() -> None:
    os.chdir(str(HERE))
    make_links()


if __name__ == "__main__":
    init_runtime()



"""
Unit test file.
"""

import unittest
from pathlib import Path

from fastled import CompileServer, Docker, Test  # type: ignore

HERE = Path(__file__).parent
PROJECT_ROOT = HERE.parent
FASTLED_SISTER_REPO = PROJECT_ROOT / ".." / "FastLED"

DEFAULT_GITHUB_URL = "https://github.com/fastled/fastled"
OUTPUT_DIR = Path(".cache/fastled")


def _enabled() -> bool:
    """Check if this system can run the tests."""
    sister_repo_does_not_exist = not FASTLED_SISTER_REPO.exists()

    if sister_repo_does_not_exist:
        print(
            f"This test is only enable when FastLED is a repo in the same directly as the project root folder: {FASTLED_SISTER_REPO} does not exist"
        )
        return False

    return Test.can_run_local_docker_tests()


class BuildDockerFromGithubTester(unittest.TestCase):

    @unittest.skipUnless(_enabled(), "Skipping test on non-Linux system on github")
    def test_build_docker_from_github(self) -> None:
        """Builds the docker file from the fastled repo."""
        url = DEFAULT_GITHUB_URL
        print("Building from github")
        server: CompileServer = Docker.spawn_server_from_github(
            url=url, output_dir=OUTPUT_DIR
        )
        self.assertTrue(server, "Failed to build docker image")
        try:
            self.assertTrue(server.ping())
        finally:
            server.stop()


if __name__ == "__main__":
    unittest.main()



"""FastLED Wasm Compiler package."""

# context
import shutil
import subprocess
import tempfile
from contextlib import contextmanager
from multiprocessing import Process
from pathlib import Path
from typing import Generator

import httpx

from .compile_server import CompileServer
from .live_client import LiveClient
from .site.build import build
from .types import BuildMode, CompileResult, CompileServerError

# IMPORTANT! There's a bug in github which will REJECT any version update
# that has any other change in the repo. Please bump the version as the
# ONLY change in a commit, or else the pypi update and the release will fail.
__version__ = "1.2.41"

DOCKER_FILE = (
    "https://raw.githubusercontent.com/zackees/fastled-wasm/refs/heads/main/Dockerfile"
)


class Api:
    @staticmethod
    def get_examples(host: str | CompileServer | None = None) -> list[str]:
        from fastled.project_init import get_examples

        if isinstance(host, CompileServer):
            host = host.url()

        return get_examples(host=host)

    @staticmethod
    def project_init(
        example=None, outputdir=None, host: str | CompileServer | None = None
    ) -> Path:
        from fastled.project_init import project_init

        if isinstance(host, CompileServer):
            host = host.url()
        return project_init(example, outputdir, host)

    @staticmethod
    def web_compile(
        directory: Path | str,
        host: str | CompileServer | None = None,
        build_mode: BuildMode = BuildMode.QUICK,
        profile: bool = False,  # When true then profile information will be enabled and included in the zip.
    ) -> CompileResult:
        from fastled.web_compile import web_compile

        if isinstance(host, CompileServer):
            host = host.url()
        if isinstance(directory, str):
            directory = Path(directory)
        out: CompileResult = web_compile(
            directory, host, build_mode=build_mode, profile=profile
        )
        return out

    @staticmethod
    def live_client(
        sketch_directory: Path,
        host: str | CompileServer | None = None,
        auto_updates: bool = True,
        auto_start=True,
        open_web_browser=True,
        keep_running=True,
        build_mode=BuildMode.QUICK,
        profile=False,
    ) -> LiveClient:
        return LiveClient(
            sketch_directory=sketch_directory,
            host=host,
            auto_updates=auto_updates,
            auto_start=auto_start,
            open_web_browser=open_web_browser,
            keep_running=keep_running,
            build_mode=build_mode,
            profile=profile,
        )

    @staticmethod
    def spawn_server(
        interactive=False,
        auto_updates=None,
        auto_start=True,
        mapped_dir: Path | None = None,  # Sketch directory.
        container_name: str | None = None,  # Specific docker container name.
        remove_previous: bool = False,
    ) -> CompileServer:
        """Uses docker to spawn a compile server from the given name."""
        from fastled.compile_server import CompileServer

        out = CompileServer(
            container_name=container_name,
            interactive=interactive,
            auto_updates=auto_updates,
            mapped_dir=mapped_dir,
            auto_start=auto_start,
            remove_previous=remove_previous,
        )
        return out

    @staticmethod
    @contextmanager
    def server(
        interactive=False,
        auto_updates=None,
        auto_start=True,
        mapped_dir: Path | None = None,  # Sketch directory.
        container_name: str | None = None,  # Specific docker container name.
        remove_previous=False,
    ) -> Generator[CompileServer, None, None]:
        server = Api.spawn_server(
            interactive=interactive,
            auto_updates=auto_updates,
            auto_start=auto_start,
            mapped_dir=mapped_dir,
            container_name=container_name,
            remove_previous=remove_previous,
        )
        try:
            yield server
        finally:
            server.stop()


class Docker:
    @staticmethod
    def is_installed() -> bool:
        from fastled.docker_manager import DockerManager

        return DockerManager.is_docker_installed()

    @staticmethod
    def is_running() -> bool:
        from fastled.docker_manager import DockerManager

        return DockerManager.is_running()

    @staticmethod
    def is_container_running(container_name: str | None = None) -> bool:
        # from fastled.docker import is_container_running
        from fastled.docker_manager import DockerManager
        from fastled.settings import CONTAINER_NAME

        docker_mgr = DockerManager()
        container_name = container_name or CONTAINER_NAME
        return docker_mgr.is_container_running(container_name)

    @staticmethod
    def purge() -> None:
        from fastled.docker_manager import DockerManager
        from fastled.settings import IMAGE_NAME

        docker_mgr = DockerManager()
        docker_mgr.purge(image_name=IMAGE_NAME)

    @staticmethod
    def spawn_server_from_github(
        url: str = "https://github.com/fastled/fastled",
        output_dir: Path | str = Path(".cache/fastled"),
    ) -> CompileServer:
        """Build the FastLED WASM compiler Docker image from a GitHub repository.

        Args:
            url: GitHub repository URL (default: https://github.com/fastled/fastled)
            output_dir: Directory to clone the repo into (default: .cache/fastled)

        Returns:
            Container name.
        """

        from fastled.docker_manager import DockerManager
        from fastled.settings import CONTAINER_NAME, IMAGE_NAME

        if isinstance(output_dir, str):
            output_dir = Path(output_dir)

        # Create output directory if it doesn't exist
        output_dir.mkdir(parents=True, exist_ok=True)

        git_dir = output_dir / ".git"
        library_properties = output_dir / "library.properties"

        git_dir_exists = git_dir.exists()
        library_properties_exists = library_properties.exists()
        library_properties_text = (
            library_properties.read_text().strip() if library_properties_exists else ""
        )

        already_exists = (
            git_dir_exists
            and library_properties_exists
            and "FastLED" in library_properties_text
        )
        if git_dir_exists and not already_exists:
            if ".cache/fastled" in str(output_dir.as_posix()):
                shutil.rmtree(output_dir)
                already_exists = False
            else:
                raise ValueError(
                    f"Output directory {output_dir} already exists but does not appear to be a FastLED repository."
                )

        # Clone or update the repository
        if already_exists:
            print(f"Updating existing repository in {output_dir}")
            # Reset local changes and move HEAD back to handle force pushes
            subprocess.run(
                ["git", "reset", "--hard", "HEAD~10"],
                cwd=output_dir,
                check=False,
                capture_output=True,  # Suppress output of reset
            )
            subprocess.run(
                ["git", "pull", "origin", "master"], cwd=output_dir, check=True
            )
        else:
            print(f"Cloning {url} into {output_dir}")
            subprocess.run(
                ["git", "clone", "--depth", "1", url, str(output_dir)], check=True
            )

        with tempfile.TemporaryDirectory() as tempdir:
            dockerfiles_dst = Path(tempdir) / "Dockerfile"
            # download the file and write it to dockerfiles_dst path
            with open(dockerfiles_dst, "wb") as f:
                with httpx.stream("GET", DOCKER_FILE) as response:
                    for chunk in response.iter_bytes():
                        f.write(chunk)

            if not dockerfiles_dst.exists():
                raise FileNotFoundError(
                    f"Dockerfile not found at {dockerfiles_dst}. "
                    "This may not be a valid FastLED repository."
                )

            docker_mgr = DockerManager()

            platform_tag = ""
            # if "arm" in docker_mgr.architecture():
            if (
                "arm"
                in subprocess.run(["uname", "-m"], capture_output=True).stdout.decode()
            ):
                platform_tag = "-arm64"

            # Build the image
            docker_mgr.build_image(
                image_name=IMAGE_NAME,
                tag="main",
                dockerfile_path=dockerfiles_dst,
                build_context=output_dir,
                build_args={"NO_PREWARM": "1"},
                platform_tag=platform_tag,
            )

        # # Run the container and return it
        # container = docker_mgr.run_container_detached(
        #     image_name=IMAGE_NAME,
        #     tag="main",
        #     container_name=CONTAINER_NAME,
        #     command=None,  # Use default command from Dockerfile
        #     volumes=None,  # No volumes needed for build
        #     ports=None,  # No ports needed for build
        #     remove_previous=True,  # Remove any existing container
        # )
        # name = container.name
        # container.stop()

        out: CompileServer = CompileServer(
            container_name=CONTAINER_NAME,
            interactive=False,
            auto_updates=False,
            mapped_dir=None,
            auto_start=True,
            remove_previous=True,
        )

        return out

    @staticmethod
    def spawn_server_from_fastled_repo(
        project_root: Path | str = Path("."),
        interactive: bool = False,
        sketch_folder: Path | None = None,
    ) -> CompileServer:
        """Build the FastLED WASM compiler Docker image, which will be tagged as "main".

        Args:
            project_root: Path to the FastLED project root directory
            platform_tag: Optional platform tag (e.g. "-arm64" for ARM builds)

        Returns:
            The string name of the docker container.
        """
        from fastled.docker_manager import DockerManager
        from fastled.settings import CONTAINER_NAME, IMAGE_NAME

        project_root = Path(project_root)
        if interactive:
            if sketch_folder is None:
                sketch_folder = project_root / "examples" / "wasm"

        if isinstance(project_root, str):
            project_root = Path(project_root)

        if DockerManager.is_docker_installed() is False:
            raise Exception("Docker is not installed.")

        docker_mgr = DockerManager()
        if DockerManager.is_running() is False:
            docker_mgr.start()

        platform_tag = ""
        # if "arm" in docker_mgr.architecture():
        if (
            "arm"
            in subprocess.run(["uname", "-m"], capture_output=True).stdout.decode()
        ):
            platform_tag = "-arm64"

        # if image exists, remove it
        docker_mgr.purge(image_name=IMAGE_NAME)

        with tempfile.TemporaryDirectory() as tempdir:
            dockerfile_dst = Path(tempdir) / "Dockerfile"

            # download the file and write it to dockerfiles_dst path
            with open(dockerfile_dst, "wb") as f:
                with httpx.stream("GET", DOCKER_FILE) as response:
                    for chunk in response.iter_bytes():
                        f.write(chunk)

            # Build the image
            docker_mgr.build_image(
                image_name=IMAGE_NAME,
                tag="main",
                dockerfile_path=dockerfile_dst,
                build_context=project_root,
                build_args={"NO_PREWARM": "1"},
                platform_tag=platform_tag,
            )

        out: CompileServer = CompileServer(
            container_name=CONTAINER_NAME,
            interactive=interactive,
            auto_updates=False,
            mapped_dir=sketch_folder,
            auto_start=True,
            remove_previous=True,
        )

        return out


class Tools:
    @staticmethod
    def string_diff(needle: str, haystack: list[str]) -> list[tuple[float, str]]:
        """Returns a sorted list with the top matches at the beginning."""
        from fastled.string_diff import string_diff

        return string_diff(needle, haystack)


class Test:
    __test__ = False  # This prevents unittest from recognizing it as a test class.

    @staticmethod
    def can_run_local_docker_tests() -> bool:
        from fastled.test.can_run_local_docker_tests import can_run_local_docker_tests

        return can_run_local_docker_tests()

    @staticmethod
    def test_examples(
        examples: list[str] | None = None, host: str | CompileServer | None = None
    ) -> dict[str, Exception]:
        from fastled.test.examples import test_examples

        if isinstance(host, CompileServer):
            host = host.url()

        return test_examples(examples=examples, host=host)

    @staticmethod
    def build_site(outputdir: Path, fast: bool | None = None, check: bool = True):
        """Builds the FastLED compiler site."""
        build(outputdir=outputdir, fast=fast, check=check)

    @staticmethod
    def spawn_http_server(
        directory: Path | str = Path("."),
        port: int | None = None,
        open_browser: bool = True,
    ) -> Process:
        from fastled.open_browser import open_browser_process

        if isinstance(directory, str):
            directory = Path(directory)
        proc: Process = open_browser_process(
            directory, port=port, open_browser=open_browser
        )
        return proc


__all__ = [
    "Api",
    "Test",
    "Build",
    "CompileServer",
    "CompileResult",
    "CompileServerError",
    "BuildMode",
]




import os
from pathlib import Path
from typing import Callable
import subprocess
import time

from compile_lock import COMPILE_LOCK

VOLUME_MAPPED_SRC = Path("/host/fastled/src")
RSYNC_DEST = Path("/js/fastled/src")
TIME_START = time.time()

def sync_src_to_target(
    src: Path, dst: Path, callback: Callable[[], None] | None = None
) -> bool:
    """Sync the volume mapped source directory to the FastLED source directory."""
    suppress_print = (
        TIME_START + 30 > time.time()
    )  # Don't print during initial volume map.
    if not src.exists():
        # Volume is not mapped in so we don't rsync it.
        print(f"Skipping rsync, as fastled src at {src} doesn't exist")
        return False
    try:
        exclude_hidden = "--exclude=.*/"  # suppresses folders like .mypy_cache/
        print("\nSyncing source directories...")
        with COMPILE_LOCK:
            # Use rsync to copy files, preserving timestamps and deleting removed files
            cp: subprocess.CompletedProcess = subprocess.run(
                ["rsync", "-av", "--info=NAME", "--delete", f"{src}/", f"{dst}/", exclude_hidden],
                check=True,
                text=True,
                capture_output=True,
            )
            if cp.returncode == 0:
                changed = False
                changed_lines: list[str] = []
                lines = cp.stdout.split("\n")
                for line in lines:
                    suffix = line.strip().split(".")[-1]
                    if suffix in ["cpp", "h", "hpp", "ino", "py", "js", "html", "css"]:
                        if not suppress_print:
                            print(f"Changed file: {line}")
                        changed = True
                        changed_lines.append(line)
                if changed:
                    if not suppress_print:
                        print(f"FastLED code had updates: {changed_lines}")
                    if callback:
                        callback()
                    return True
                print("Source directory synced successfully with no changes")
                return False
            else:
                print(f"Error syncing directories: {cp.stdout}\n\n{cp.stderr}")
                return False

    except subprocess.CalledProcessError as e:
        print(f"Error syncing directories: {e.stdout}\n\n{e.stderr}")
    except Exception as e:
        print(f"Error syncing directories: {e}")
    return False



def sync_source_directory_if_volume_is_mapped(
    callback: Callable[[], None] | None = None
) -> bool:
    """Sync the volume mapped source directory to the FastLED source directory."""
    if not VOLUME_MAPPED_SRC.exists():
        # Volume is not mapped in so we don't rsync it.
        print("Skipping rsync, as fastled src volume not mapped")
        return False
    print("Syncing source directories because host is mapped in")
    return sync_src_to_target(VOLUME_MAPPED_SRC, RSYNC_DEST, callback=callback)


# A compilation script specific to fastled's docker compiler.
# This script will pull the users code from a mapped directory,
# then do some processing to convert the *.ino files to *.cpp and
# insert certain headers like "Arduino.h" (pointing to a fake implementation).
# After this, the code is compiled, and the output files are copied back
# to the users mapped directory in the fastled_js folder.
# There are a few assumptions for this script:
# 1. The mapped directory will contain only one directory with the users code, this is
#    enforced by the script that sets up the docker container.
# 2. The docker container has installed compiler dependencies in the /js directory.

import argparse
import hashlib
import json
import os
import re
import shutil
import subprocess
import sys
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import List


@dataclass
class DateLine:
    dt: datetime
    line: str


class BuildMode(Enum):
    DEBUG = "DEBUG"
    QUICK = "QUICK"
    RELEASE = "RELEASE"

    @classmethod
    def from_string(cls, mode_str: str) -> "BuildMode":
        try:
            return cls[mode_str.upper()]
        except KeyError:
            valid_modes = [mode.name for mode in cls]
            raise ValueError(f"BUILD_MODE must be one of {valid_modes}, got {mode_str}")


@dataclass
class SyntaxCheckResult:
    file_path: Path
    is_valid: bool
    message: str


_CHECK_SYNTAX = False
_COMPILER_PATH = "em++"

JS_DIR = Path("/js")
FASTLED_DIR = JS_DIR / "fastled"
FASTLED_SRC = FASTLED_DIR / "src"
# FASTLED_SRC_PLATFORMS = FASTLED_SRC / "platforms"
# FASTLED_SRC_PLATFORMS_WASM = FASTLED_SRC_PLATFORMS / "wasm"
# FASTLED_SRC_PLATFORMS_WASM_COMPILER = FASTLED_SRC_PLATFORMS_WASM / "compiler"
FASTLED_SRC_PLATFORMS_WASM_COMPILER = JS_DIR / "compiler"


JS_SRC = JS_DIR / "src"

FASTLED_DIR = JS_DIR / "fastled"
# FASTLED_SRC_DIR = FASTLED_DIR / "src"
# FASTLED_PLATFORMS_DIR = FASTLED_SRC_DIR / "platforms"
# FASTLED_WASM_DIR = FASTLED_PLATFORMS_DIR / "wasm"
FASTLED_COMPILER_DIR = JS_DIR / "compiler"
FASTLED_MODULES_DIR = FASTLED_COMPILER_DIR / "modules"

PIO_BUILD_DIR = JS_DIR / ".pio/build"
ARDUINO_H_SRC = JS_DIR / "Arduino.h"
INDEX_HTML_SRC = FASTLED_COMPILER_DIR / "index.html"
INDEX_CSS_SRC = FASTLED_COMPILER_DIR / "index.css"
INDEX_JS_SRC = FASTLED_COMPILER_DIR / "index.js"


WASM_COMPILER_SETTTINGS = JS_DIR / "wasm_compiler_flags.py"
OUTPUT_FILES = ["fastled.js", "fastled.wasm"]
HEADERS_TO_INSERT = ["#include <Arduino.h>", '#include "platforms/wasm/js.h"']
FILE_EXTENSIONS = [".ino", ".h", ".hpp", ".cpp"]
MAX_COMPILE_ATTEMPTS = 1  # Occasionally the compiler fails for unknown reasons, but disabled because it increases the build time on failure.
FASTLED_OUTPUT_DIR_NAME = "fastled_js"




def copy_files(src_dir: Path, js_src: Path) -> None:
    print("Copying files from mapped directory to container...")
    for item in src_dir.iterdir():
        if item.is_dir():
            print(f"Copying directory: {item}")
            shutil.copytree(item, js_src / item.name, dirs_exist_ok=True)
        else:
            print(f"Copying file: {item}")
            shutil.copy2(item, js_src / item.name)


def compile(
    js_dir: Path, build_mode: BuildMode, auto_clean: bool, no_platformio: bool
) -> int:
    print("Starting compilation process...")
    max_attempts = 1
    env = os.environ.copy()
    env["BUILD_MODE"] = build_mode.name
    print(f"Build mode: {build_mode.name}")
    cmd_list: list[str] = []
    if no_platformio:
        # execute build_archive.syh
        cmd_list = [
            "/bin/bash",
            "-c",
            "/js/build_fast.sh",
        ]
    else:
        cmd_list.extend(["pio", "run"])
        if not auto_clean:
            cmd_list.append("--disable-auto-clean")

    def _open_process(cmd_list: list[str] = cmd_list) -> subprocess.Popen:
        out = subprocess.Popen(
            cmd_list,
            cwd=js_dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            env=env,
        )
        return out

    output_lines = []
    for attempt in range(1, max_attempts + 1):
        try:
            print(f"Attempting compilation (attempt {attempt}/{max_attempts})...")
            process = _open_process()
            assert process.stdout is not None
            for line in process.stdout:
                processed_line = line.replace("fastled/src", "src")
                timestamped_line = _timestamp_output(processed_line)
                output_lines.append(timestamped_line)
            process.wait()
            relative_output = _make_timestamps_relative("\n".join(output_lines))
            print(relative_output)
            if process.returncode == 0:
                print(f"Compilation successful on attempt {attempt}")
                return 0
            else:
                raise subprocess.CalledProcessError(process.returncode, ["pio", "run"])
        except subprocess.CalledProcessError:
            print(f"Compilation failed on attempt {attempt}")
            if attempt == max_attempts:
                print("Max attempts reached. Compilation failed.")
                return 1
            print("Retrying...")
    return 1


def insert_header(file: Path) -> None:
    print(f"Inserting header in file: {file}")
    with open(file, "r") as f:
        content = f.read()

    # Remove existing includes
    for header in HEADERS_TO_INSERT:
        content = re.sub(
            rf"^.*{re.escape(header)}.*\n", "", content, flags=re.MULTILINE
        )

    # Remove both versions of Arduino.h include
    arduino_pattern = r'^\s*#\s*include\s*[<"]Arduino\.h[>"]\s*.*\n'
    content = re.sub(arduino_pattern, "", content, flags=re.MULTILINE)

    # Add new headers at the beginning
    content = "\n".join(HEADERS_TO_INSERT) + "\n" + content

    with open(file, "w") as f:
        f.write(content)
    print(f"Processed: {file}")


def transform_to_cpp(src_dir: Path) -> None:
    print("Transforming files to cpp...")
    ino_files = list(src_dir.glob("*.ino"))

    if ino_files:
        ino_file = ino_files[0]
        print(f"Found .ino file: {ino_file}")
        main_cpp = src_dir / "main.cpp"
        if main_cpp.exists():
            print("main.cpp already exists, renaming to main2.hpp")
            main_cpp.rename(src_dir / "main2.hpp")

        new_cpp_file = ino_file.with_suffix(".ino.cpp")
        print(f"Renaming {ino_file} to {new_cpp_file.name}")
        ino_file.rename(new_cpp_file)

        if (src_dir / "main2.hpp").exists():
            print(f"Including main2.hpp in {new_cpp_file.name}")
            with open(new_cpp_file, "a") as f:
                f.write('#include "main2.hpp"\n')


def insert_headers(
    src_dir: Path, exclusion_folders: List[Path], file_extensions: List[str]
) -> None:
    print("Inserting headers in source files...")
    for file in src_dir.rglob("*"):
        if (
            file.suffix in file_extensions
            and not any(folder in file.parents for folder in exclusion_folders)
            and file.name != "Arduino.h"
        ):
            insert_header(file)


def process_ino_files(src_dir: Path) -> None:
    transform_to_cpp(src_dir)
    exclusion_folders: List[Path] = []
    insert_headers(src_dir, exclusion_folders, FILE_EXTENSIONS)
    print("Transform to cpp and insert header operations completed.")


def check_syntax_with_gcc(file_path, gcc_path="gcc"):
    """
    Perform syntax checking on a C or C++ source file using GCC.

    Parameters:
        file_path (str): Path to the source file to check.
        gcc_path (str): Path to the GCC executable (default is 'gcc').

    Returns:
        bool: True if syntax is correct, False otherwise.
        str: Output or error message from GCC.
    """
    try:
        # Run GCC with -fsyntax-only flag for syntax checking
        cmd_list = [
            gcc_path,
            "-fsyntax-only",
            "-std=gnu++20",
            "-fpermissive",
            "-Wno-everything",  # Suppress all warnings
            "-I",
            "/js/src/",  # Add /js/src/ to the include path
            "-I",
            "/js/fastled/src/",  # Add /js/fastled/src/ to the include path
            "-I",
            "/emsdk/upstream/emscripten/system/include",
            file_path,
        ]
        cmd_str = subprocess.list2cmdline(cmd_list)
        print(f"Running command: {cmd_str}")
        result = subprocess.run(
            cmd_list,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        # Check the return code to determine if syntax is valid
        if result.returncode == 0:
            return True, "Syntax check passed successfully."
        else:
            return False, result.stderr
    except FileNotFoundError:
        return False, f"GCC not found at {gcc_path}."
    except Exception as e:
        return False, str(e)


def check_syntax(
    directory_path: Path, gcc_path: str = "gcc"
) -> list[SyntaxCheckResult]:
    # os walk
    out: list[SyntaxCheckResult] = []
    exclusion_list = set("fastled_js")
    for root, dirs, files in os.walk(directory_path):
        # if sub directory is in exclusion list, skip
        dirs[:] = [d for d in dirs if d not in exclusion_list]
        for file in files:
            if file.endswith(".cpp") or file.endswith(".ino"):
                file_path = os.path.join(root, file)
                is_valid, message = check_syntax_with_gcc(file_path, gcc_path)
                if not is_valid:
                    print(f"Syntax check failed for file: {file_path}")
                    print(f"Error message: {message}")
                    out.append(
                        SyntaxCheckResult(
                            file_path=Path(file_path), is_valid=False, message=message
                        )
                    )
                else:
                    print(f"Syntax check passed for file: {file_path}")
                    out.append(
                        SyntaxCheckResult(
                            file_path=Path(file_path),
                            is_valid=True,
                            message="Syntax check passed successfully.",
                        )
                    )
    return out


def _make_timestamps_relative(stdout: str) -> str:
    def parse(line: str) -> DateLine:
        parts = line.split(" ")
        if len(parts) < 2:
            raise ValueError(f"Invalid line: {line}")

        date_str, time_str = parts[:2]
        rest = " ".join(parts[2:])
        # Parse with microsecond precision
        dt = datetime.strptime(f"{date_str} {time_str}", "%Y-%m-%d %H:%M:%S.%f")
        return DateLine(dt, rest)

    lines = stdout.split("\n")
    if not lines:
        return stdout
    parsed: list[DateLine] = []
    for line in lines:
        if not line.strip():  # Skip empty lines
            continue
        try:
            parsed.append(parse(line))
        except ValueError:
            print(f"Failed to parse line: {line}")
            continue

    if not parsed:
        return stdout

    outlines: list[str] = []
    start_time = parsed[0].dt

    # Calculate relative times with
    for p in parsed:
        delta = p.dt - start_time
        seconds = delta.total_seconds()
        line_str = f"{seconds:3.2f} {p.line}"
        outlines.append(line_str)

    return "\n".join(outlines)


def _timestamp_output(line: str) -> str:
    now = datetime.now()
    timestamp = now.strftime("%Y-%m-%d %H:%M:%S.%f")
    return f"{timestamp} {line.rstrip()}"


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Compile FastLED for WASM")
    parser.add_argument(
        "--mapped-dir",
        type=Path,
        default="/mapped",
        help="Directory containing source files (default: /mapped)",
    )
    parser.add_argument(
        "--keep-files", action="store_true", help="Keep source files after compilation"
    )
    parser.add_argument(
        "--only-copy",
        action="store_true",
        help="Only copy files from mapped directory to container",
    )
    parser.add_argument(
        "--only-insert-header",
        action="store_true",
        help="Only insert headers in source files",
    )
    parser.add_argument(
        "--only-compile", action="store_true", help="Only compile the project"
    )
    parser.add_argument(
        "--profile",
        action="store_true",
        help="Enable profiling for compilation to see what's taking so long.",
    )

    parser.add_argument(
        "--disable-auto-clean",
        action="store_true",
        help="Massaive speed improvement to not have to rebuild everything, but flakes out sometimes.",
        default=os.getenv("DISABLE_AUTO_CLEAN", "0") == "1",
    )
    parser.add_argument(
        "--no-platformio",
        action="store_true",
        help="Don't use platformio to compile the project, use the new system of direct emcc calls.",
    )
    # Add mutually exclusive build mode group
    build_mode = parser.add_mutually_exclusive_group()
    build_mode.add_argument("--debug", action="store_true", help="Build in debug mode")
    build_mode.add_argument(
        "--quick",
        action="store_true",
        default=True,
        help="Build in quick mode (default)",
    )
    build_mode.add_argument(
        "--release", action="store_true", help="Build in release mode"
    )

    return parser.parse_args()


def find_project_dir(mapped_dir: Path) -> Path:
    mapped_dirs: List[Path] = list(mapped_dir.iterdir())
    if len(mapped_dirs) > 1:
        raise ValueError(
            f"Error: More than one directory found in {mapped_dir}, which are {mapped_dirs}"
        )

    src_dir: Path = mapped_dirs[0]
    return src_dir


def process_compile(
    js_dir: Path, build_mode: BuildMode, auto_clean: bool, no_platformio: bool
) -> None:
    print("Starting compilation...")
    rtn = compile(js_dir, build_mode, auto_clean, no_platformio=no_platformio)
    print(f"Compilation return code: {rtn}")
    if rtn != 0:
        print("Compilation failed.")
        raise RuntimeError("Compilation failed.")
    print("Compilation successful.")


def cleanup(args: argparse.Namespace, js_src: Path) -> None:
    if not args.keep_files and not (args.only_copy or args.only_insert_header):
        print("Removing temporary source files")
        shutil.rmtree(js_src)
    else:
        print("Keeping temporary source files")


def hash_file(file_path: Path) -> str:
    hasher = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


def main() -> int:



    # assert JS_DIR.exists()
    # assert ARDUINO_H_SRC.exists()
    # assert INDEX_HTML_SRC.exists()
    # assert INDEX_CSS_SRC.exists(), f"Index CSS not found at {INDEX_CSS_SRC}"
    # assert INDEX_JS_SRC.exists()
    # assert WASM_COMPILER_SETTTINGS.exists()
    # assert FASTLED_SRC_PLATFORMS_WASM_COMPILER.exists()
    # assert JS_DIR.exists(), f"JS_DIR does not exist: {JS_DIR}"
    # assert ARDUINO_H_SRC.exists(), f"ARDUINO_H_SRC does not exist: {ARDUINO_H_SRC}"
    # assert INDEX_HTML_SRC.exists(), f"INDEX_HTML_SRC does not exist: {INDEX_HTML_SRC}"
    # assert INDEX_CSS_SRC.exists(), f"INDEX_CSS_SRC does not exist: {INDEX_CSS_SRC}"
    # assert INDEX_JS_SRC.exists(), f"INDEX_JS_SRC does not exist: {INDEX_JS_SRC}"
    # assert (
    #     WASM_COMPILER_SETTTINGS.exists()
    # ), f"WASM_COMPILER_SETTTINGS does not exist: {WASM_COMPILER_SETTTINGS}"

    check_paths: list[Path] = [
        JS_DIR,
        ARDUINO_H_SRC,
        INDEX_HTML_SRC,
        INDEX_CSS_SRC,
        INDEX_JS_SRC,
        WASM_COMPILER_SETTTINGS,
        FASTLED_SRC_PLATFORMS_WASM_COMPILER,
    ]

    missing_paths = [p for p in check_paths if not p.exists()]
    if missing_paths:
        print("The following paths are missing:")
        for p in missing_paths:
            print(p)
        missing_paths_str = ",".join(str(p.as_posix()) for p in missing_paths)
        raise FileNotFoundError(f"Missing required paths: {missing_paths_str}")

    print("Starting FastLED WASM compilation script...")
    args = parse_args()
    print(f"Keep files flag: {args.keep_files}")
    print(f"Using mapped directory: {args.mapped_dir}")

    if args.profile:
        print("Enabling profiling for compilation.")
        # Profile linking
        os.environ["EMPROFILE"] = "2"

    try:
        if JS_SRC.exists():
            shutil.rmtree(JS_SRC)
        JS_SRC.mkdir(parents=True, exist_ok=True)
        src_dir = find_project_dir(args.mapped_dir)

        any_only_flags = args.only_copy or args.only_insert_header or args.only_compile

        do_copy = not any_only_flags or args.only_copy
        do_insert_header = not any_only_flags or args.only_insert_header
        do_compile = not any_only_flags or args.only_compile

        if do_copy:
            copy_files(src_dir, JS_SRC)
            print("Copying Arduino.h to src/Arduino.h")
            shutil.copy(ARDUINO_H_SRC, JS_SRC / "Arduino.h")
            if args.only_copy:
                return 0

        if do_insert_header:
            process_ino_files(JS_SRC)
            if args.only_insert_header:
                print("Transform to cpp and insert header operations completed.")
                return 0

        if _CHECK_SYNTAX:
            print("Performing syntax check...")
            syntax_results = check_syntax(
                directory_path=JS_SRC, gcc_path=_COMPILER_PATH
            )
            failed_checks = [r for r in syntax_results if not r.is_valid]
            if failed_checks:
                print("\nSyntax check failed!")
                for result in failed_checks:
                    print(f"\nFile: {result.file_path}")
                    print(f"Error: {result.message}")
                return 1
            print("Syntax check passed for all files.")

        no_platformio: bool = args.no_platformio

        if do_compile:
            try:
                # Determine build mode from args
                if args.debug:
                    build_mode = BuildMode.DEBUG
                elif args.release:
                    build_mode = BuildMode.RELEASE
                else:
                    # Default to QUICK mode if neither debug nor release specified
                    build_mode = BuildMode.QUICK

                process_compile(
                    js_dir=JS_DIR,
                    build_mode=build_mode,
                    auto_clean=not args.disable_auto_clean,
                    no_platformio=no_platformio,
                )
            except Exception as e:
                print(f"Error: {str(e)}")
                return 1

            def _get_build_dir_platformio() -> Path:
                build_dirs = [d for d in PIO_BUILD_DIR.iterdir() if d.is_dir()]
                if len(build_dirs) != 1:
                    raise RuntimeError(
                        f"Expected exactly one build directory in {PIO_BUILD_DIR}, found {len(build_dirs)}: {build_dirs}"
                    )
                build_dir: Path = build_dirs[0]
                return build_dir

            def _get_build_dir_cmake() -> Path:
                return JS_DIR / "build"

            if no_platformio:
                build_dir = _get_build_dir_cmake()
            else:
                build_dir = _get_build_dir_platformio()

            print("Copying output files...")
            fastled_js_dir: Path = src_dir / FASTLED_OUTPUT_DIR_NAME
            fastled_js_dir.mkdir(parents=True, exist_ok=True)

            for file in ["fastled.js", "fastled.wasm"]:
                _src = build_dir / file
                _dst = fastled_js_dir / file
                print(f"Copying {_src} to {_dst}")
                shutil.copy2(_src, _dst)

            print(f"Copying {INDEX_HTML_SRC} to output directory")
            shutil.copy2(INDEX_HTML_SRC, fastled_js_dir / "index.html")
            print(f"Copying {INDEX_CSS_SRC} to output directory")
            shutil.copy2(INDEX_CSS_SRC, fastled_js_dir / "index.css")

            # copy all js files in FASTLED_COMPILER_DIR to output directory
            Path(fastled_js_dir / "modules").mkdir(parents=True, exist_ok=True)
            for _file in FASTLED_MODULES_DIR.iterdir():
                if _file.suffix == ".js":
                    print(f"Copying {_file} to output directory")
                    shutil.copy2(_file, fastled_js_dir / "modules" / _file.name)

            fastled_js_mem = build_dir / "fastled.js.mem"
            fastled_wasm_map = build_dir / "fastled.wasm.map"
            fastled_js_symbols = build_dir / "fastled.js.symbols"
            if fastled_js_mem.exists():
                print(f"Copying {fastled_js_mem} to output directory")
                shutil.copy2(fastled_js_mem, fastled_js_dir / fastled_js_mem.name)
            if fastled_wasm_map.exists():
                print(f"Copying {fastled_wasm_map} to output directory")
                shutil.copy2(fastled_wasm_map, fastled_js_dir / fastled_wasm_map.name)
            if fastled_js_symbols.exists():
                print(f"Copying {fastled_js_symbols} to output directory")
                shutil.copy2(
                    fastled_js_symbols, fastled_js_dir / fastled_js_symbols.name
                )
            print("Copying index.js to output directory")
            shutil.copy2(INDEX_JS_SRC, fastled_js_dir / "index.js")
            optional_input_data_dir = src_dir / "data"
            output_data_dir = fastled_js_dir / optional_input_data_dir.name

            # Handle data directory if it exists
            manifest: list[dict] = []
            if optional_input_data_dir.exists():
                # Clean up existing output data directory
                if output_data_dir.exists():
                    for _file in output_data_dir.iterdir():
                        _file.unlink()

                # Create output data directory and copy files
                output_data_dir.mkdir(parents=True, exist_ok=True)
                for _file in optional_input_data_dir.iterdir():
                    if _file.is_file():  # Only copy files, not directories
                        filename: str = _file.name
                        if filename.endswith(".embedded.json"):
                            print("Embedding data file")
                            filename_no_embedded = filename.replace(
                                ".embedded.json", ""
                            )
                            # read json file
                            with open(_file, "r") as f:
                                data = json.load(f)
                            hash_value = data["hash"]
                            size = data["size"]
                            manifest.append(
                                {
                                    "name": filename_no_embedded,
                                    "path": f"data/{filename_no_embedded}",
                                    "size": size,
                                    "hash": hash_value,
                                }
                            )
                        else:
                            print(f"Copying {_file.name} -> {output_data_dir}")
                            shutil.copy2(_file, output_data_dir / _file.name)
                            hash = hash_file(_file)
                            manifest.append(
                                {
                                    "name": _file.name,
                                    "path": f"data/{_file.name}",
                                    "size": _file.stat().st_size,
                                    "hash": hash,
                                }
                            )

            # Write manifest file even if empty
            print("Writing manifest files.json")
            manifest_json_str = json.dumps(manifest, indent=2, sort_keys=True)
            with open(fastled_js_dir / "files.json", "w") as f:
                f.write(manifest_json_str)
        cleanup(args, JS_SRC)

        print("Compilation process completed successfully")
        return 0

    except Exception as e:
        import traceback

        stacktrace = traceback.format_exc()
        print(stacktrace)
        print(f"Error: {str(e)}")
        return 1


if __name__ == "__main__":
    sys.exit(main())



import argparse
import subprocess
import os
import sys
from typing import Tuple
import warnings
from pathlib import Path

from code_sync import sync_source_directory_if_volume_is_mapped

_PORT = os.environ.get("PORT", 80)

_CHOICES = [
    "compile",
    "server"
]

HERE = Path(__file__).parent

# if [ "$RENDER" != "true" ]; then
#   echo "Skipping finalprewarm..."
#   exit 0
# fi

# git_path=/git/fastled
# fastled_path=/js/fastled

# # update the fastled git repo
# cd $git_path

# git fetch origin
# git reset --hard origin/master
# #  ["rsync", "-av", "--info=NAME", "--delete", f"{src}/", f"{dst}/"],

# cd /js

# rsync -av --info=NAME --delete "$git_path/" "$fastled_path/"  --exclude ".git"

def _update_fastled() -> None:
    # NOT ENABLED YET
    if True:
        return
    is_render = os.environ.get("RENDER", "false") == "true"
    if not is_render:
        print("Skipping finalprewarm...")
        return
    git_path = "/git/fastled"
    fastled_path = "/js/fastled"
    subprocess.run(["git", "fetch", "origin"], cwd=git_path)
    subprocess.run(["git", "reset", "--hard", "origin/master"], cwd=git_path)
    subprocess.run(["rsync", "-av", "--info=NAME", "--delete", f"{git_path}/", f"{fastled_path}/", "--exclude", ".git"], cwd="/js")




def _parse_args() -> Tuple[argparse.Namespace, list[str]]:
    parser = argparse.ArgumentParser(description="Run compile.py with additional arguments")
    parser.add_argument("mode", help="Which mode does this script run in", choices=_CHOICES)
    return parser.parse_known_args()

def _run_server(unknown_args: list[str]) -> int:
    env = os.environ.copy()
    if "--disable-auto-clean" in unknown_args:
        env["DISABLE_AUTO_CLEAN"] = "1"
        unknown_args.remove("--disable-auto-clean")
    if "--allow-shutdown" in unknown_args:
        env["ALLOW_SHUTDOWN"] = "1"
        unknown_args.remove("--allow-shutdown")
    if "--no-auto-update" in unknown_args:
        env["NO_AUTO_UPDATE"] = "1"
        unknown_args.remove("--no-auto-update")
    if "--no-sketch-cache" in unknown_args:
        env["NO_SKETCH_CACHE"] = "1"
        unknown_args.remove("--no-sketch-cache")
    if unknown_args:
        warnings.warn(f"Unknown arguments: {unknown_args}")
        unknown_args = []
    cmd_list = [
        "uvicorn",
        "server:app",
        "--host",
        "0.0.0.0",
        "--workers",
        "1",
        "--port",
        f"{_PORT}"
    ]
    cp: subprocess.CompletedProcess = subprocess.run(cmd_list, cwd=str(HERE), env=env)
    return cp.returncode

def _run_compile(unknown_args: list[str]) -> int:

    # Construct the command to call compile.py with unknown arguments
    command = [sys.executable, 'compile.py'] + unknown_args

    # Call compile.py with the unknown arguments
    result = subprocess.run(command, text=True, cwd=str(HERE))

    # Print the output from compile.py
    #print(result.stdout)
    #if result.stderr:
    #    print(result.stderr, file=sys.stderr)
    return result.returncode

def main() -> int:
    print("Running...")
    args, unknown_args = _parse_args()
    _update_fastled()
    sync_source_directory_if_volume_is_mapped()

    try:
        if args.mode == "compile":
            rtn = _run_compile(unknown_args)
            return rtn
        elif args.mode == "server":
            rtn = _run_server(unknown_args)
            return rtn
        raise ValueError(f"Unknown mode: {args.mode}")
    except KeyboardInterrupt:
        print("Exiting...")
        return 1


if __name__ == "__main__":
    sys.exit(main())



"""
Uses the latest wasm compiler image to compile the FastLED sketch.
"""

import sys
import time
from pathlib import Path

from fastled.client_server import run_client_server
from fastled.compile_server import CompileServer
from fastled.filewatcher import file_watcher_set
from fastled.parse_args import Args, parse_args


def run_server(args: Args) -> int:
    interactive = args.interactive
    auto_update = args.auto_update
    mapped_dir = Path(args.directory).absolute() if args.directory else None
    if interactive and mapped_dir is None:
        print("Select a sketch when you enter interactive mode.")
        return 1
    compile_server = CompileServer(
        interactive=interactive,
        auto_updates=auto_update,
        mapped_dir=mapped_dir,
        auto_start=True,
    )

    if not interactive:
        print(f"Server started at {compile_server.url()}")
    try:
        while True:
            if not compile_server.process_running():
                print("Server process is not running. Exiting...")
                return 1
            time.sleep(0.1)
    except KeyboardInterrupt:
        print("\nExiting from server...")
        return 1
    finally:
        compile_server.stop()
    return 0


def main() -> int:
    args = parse_args()
    interactive: bool = args.interactive
    has_server = args.server
    update: bool = args.update
    build: bool = args.build
    just_compile: bool = args.just_compile
    # directory: Path | None = Path(args.directory).absolute() if args.directory else None
    directory: Path | None = Path(args.directory) if args.directory else None

    if update:
        # Force auto_update to ensure update check happens
        compile_server = CompileServer(interactive=False, auto_updates=True)
        compile_server.stop()
        print("Finished updating.")
        return 0

    if build:
        file_watcher_set(False)
        project_root = Path(".").absolute()
        print(f"Building Docker image at {project_root}")
        from fastled import Api, Docker

        server = Docker.spawn_server_from_fastled_repo(
            project_root=project_root,
            interactive=interactive,
            sketch_folder=directory,
        )
        assert isinstance(server, CompileServer)

        try:
            if interactive:
                server.stop()
                return 0
            print(f"Built Docker image: {server.name}")
            if not directory:
                if not directory:
                    print("No directory specified")
                server.stop()
                return 0

            print("Running server")

            with Api.live_client(
                auto_updates=False,
                sketch_directory=directory,
                host=server,
                auto_start=True,
                keep_running=not just_compile,
            ) as _:
                while True:
                    time.sleep(0.2)  # wait for user to exit
        except KeyboardInterrupt:
            print("\nExiting from client...")
            server.stop()
            return 1

    if has_server:
        print("Running in server only mode.")
        return run_server(args)
    else:
        print("Running in client/server mode.")
        return run_client_server(args)


if __name__ == "__main__":
    # Note that the entry point for the exe is in cli.py
    try:
        sys.argv.append("-b")
        # sys.argv.append("examples/wasm")
        # sys.argv.append()
        import os

        os.chdir("../fastled")
        sys.argv.append("examples/wasm")
        sys.exit(main())
    except KeyboardInterrupt:
        print("\nExiting from main...")
        sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)



import subprocess
import sys
import time
import webbrowser
from multiprocessing import Process
from pathlib import Path

DEFAULT_PORT = 8089  # different than live version.

PYTHON_EXE = sys.executable


def open_http_server_subprocess(
    fastled_js: Path, port: int, open_browser: bool
) -> None:
    """Start livereload server in the fastled_js directory and return the process"""
    import shutil

    try:
        if shutil.which("live-server") is not None:
            cmd = [
                "live-server",
                f"--port={port}",
                "--host=localhost",
                ".",
            ]
            if not open_browser:
                cmd.append("--no-browser")
            subprocess.run(cmd, shell=True, cwd=fastled_js)
            return

        cmd = [
            PYTHON_EXE,
            "-m",
            "fastled.open_browser2",
            str(fastled_js),
            "--port",
            str(port),
        ]
        # return subprocess.Popen(cmd)  # type ignore
        # pipe stderr and stdout to null
        subprocess.run(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )  # type ignore
    except KeyboardInterrupt:
        print("Exiting from server...")
        sys.exit(0)


def is_port_free(port: int) -> bool:
    """Check if a port is free"""
    import httpx

    try:
        response = httpx.get(f"http://localhost:{port}", timeout=1)
        response.raise_for_status()
        return False
    except (httpx.HTTPError, httpx.ConnectError):
        return True


def find_free_port(start_port: int) -> int:
    """Find a free port starting at start_port"""
    for port in range(start_port, start_port + 100):
        if is_port_free(port):
            print(f"Found free port: {port}")
            return port
        else:
            print(f"Port {port} is in use, finding next")
    raise ValueError("Could not find a free port")


def wait_for_server(port: int, timeout: int = 10) -> None:
    """Wait for the server to start."""
    from httpx import get

    future_time = time.time() + timeout
    while future_time > time.time():
        try:
            response = get(f"http://localhost:{port}", timeout=1)
            if response.status_code == 200:
                return
        except Exception:
            continue
    raise TimeoutError("Could not connect to server")


def _background_npm_install_live_server() -> None:
    import shutil
    import time

    if shutil.which("npm") is None:
        return

    if shutil.which("live-server") is not None:
        return

    time.sleep(3)
    subprocess.run(
        ["npm", "install", "-g", "live-server"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def open_browser_process(
    fastled_js: Path, port: int | None = None, open_browser: bool = True
) -> Process:
    import shutil

    """Start livereload server in the fastled_js directory and return the process"""
    if port is not None:
        if not is_port_free(port):
            raise ValueError(f"Port {port} was specified but in use")
    else:
        port = find_free_port(DEFAULT_PORT)
    out: Process = Process(
        target=open_http_server_subprocess,
        args=(fastled_js, port, False),
        daemon=True,
    )
    out.start()
    wait_for_server(port)
    if open_browser:
        print(f"Opening browser to http://localhost:{port}")
        webbrowser.open(url=f"http://localhost:{port}", new=1, autoraise=True)

    # start a deamon thread to install live-server
    if shutil.which("live-server") is None:
        import threading

        t = threading.Thread(target=_background_npm_install_live_server)
        t.daemon = True
        t.start()
    return out


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Open a browser to the fastled_js directory"
    )
    parser.add_argument(
        "fastled_js",
        type=Path,
        help="Path to the fastled_js directory",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=DEFAULT_PORT,
        help=f"Port to run the server on (default: {DEFAULT_PORT})",
    )
    args = parser.parse_args()

    proc = open_browser_process(args.fastled_js, args.port, open_browser=True)
    proc.join()



"""
New abstraction for Docker management with improved Ctrl+C handling.
"""

import _thread
import os
import platform
import subprocess
import sys
import threading
import time
import traceback
import warnings
from datetime import datetime, timezone
from pathlib import Path

import docker
from appdirs import user_data_dir
from disklru import DiskLRUCache
from docker.client import DockerClient
from docker.models.containers import Container
from docker.models.images import Image
from filelock import FileLock

from fastled.spinner import Spinner

CONFIG_DIR = Path(user_data_dir("fastled", "fastled"))
CONFIG_DIR.mkdir(parents=True, exist_ok=True)
DB_FILE = CONFIG_DIR / "db.db"
DISK_CACHE = DiskLRUCache(str(DB_FILE), 10)
_IS_GITHUB = "GITHUB_ACTIONS" in os.environ


# Docker uses datetimes in UTC but without the timezone info. If we pass in a tz
# then it will throw an exception.
def _utc_now_no_tz() -> datetime:
    now = datetime.now(timezone.utc)
    return now.replace(tzinfo=None)


def _win32_docker_location() -> str | None:
    home_dir = Path.home()
    out = [
        "C:\\Program Files\\Docker\\Docker\\Docker Desktop.exe",
        f"{home_dir}\\AppData\\Local\\Docker\\Docker Desktop.exe",
    ]
    for loc in out:
        if Path(loc).exists():
            return loc
    return None


def get_lock(image_name: str) -> FileLock:
    """Get the file lock for this DockerManager instance."""
    lock_file = CONFIG_DIR / f"{image_name}.lock"
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    print(CONFIG_DIR)
    if not lock_file.parent.exists():
        lock_file.parent.mkdir(parents=True, exist_ok=True)
    return FileLock(str(lock_file))


class RunningContainer:
    def __init__(self, container, first_run=False):
        self.container = container
        self.first_run = first_run
        self.running = True
        self.thread = threading.Thread(target=self._log_monitor)
        self.thread.daemon = True
        self.thread.start()

    def _log_monitor(self):
        from_date = _utc_now_no_tz() if not self.first_run else None
        to_date = _utc_now_no_tz()

        while self.running:
            try:
                for log in self.container.logs(
                    follow=False, since=from_date, until=to_date, stream=True
                ):
                    print(log.decode("utf-8"), end="")
                time.sleep(0.1)
                from_date = to_date
                to_date = _utc_now_no_tz()
            except KeyboardInterrupt:
                print("Monitoring logs interrupted by user.")
                _thread.interrupt_main()
                break
            except Exception as e:
                print(f"Error monitoring logs: {e}")
                break

    def detach(self) -> None:
        """Stop monitoring the container logs"""
        self.running = False
        self.thread.join()


class DockerManager:
    def __init__(self) -> None:
        from docker.errors import DockerException

        try:
            self._client: DockerClient | None = None
            self.first_run = False
        except DockerException as e:
            stack = traceback.format_exc()
            warnings.warn(f"Error initializing Docker client: {e}\n{stack}")
            raise

    @property
    def client(self) -> DockerClient:
        if self._client is None:
            self._client = docker.from_env()
        return self._client

    @staticmethod
    def is_docker_installed() -> bool:
        """Check if Docker is installed on the system."""
        try:
            subprocess.run(["docker", "--version"], capture_output=True, check=True)
            print("Docker is installed.")
            return True
        except subprocess.CalledProcessError as e:
            print(f"Docker command failed: {str(e)}")
            return False
        except FileNotFoundError:
            print("Docker is not installed.")
            return False

    @staticmethod
    def ensure_linux_containers_for_windows() -> bool:
        """Ensure Docker is using Linux containers on Windows."""
        if sys.platform != "win32":
            return True  # Only needed on Windows

        try:
            # Check if we're already in Linux container mode
            result = subprocess.run(
                ["docker", "info"], capture_output=True, text=True, check=True
            )

            if "linux" in result.stdout.lower():
                print("Already using Linux containers")
                return True

            if not _IS_GITHUB:
                answer = (
                    input(
                        "\nDocker on Windows must be in linux mode, this is a global change, switch? [y/n]"
                    )
                    .strip()
                    .lower()[:1]
                )
                if answer != "y":
                    return False

            print("Switching to Linux containers...")
            warnings.warn("Switching Docker to use Linux container context...")

            # Explicitly specify the Linux container context
            linux_context = "desktop-linux"
            subprocess.run(
                ["cmd", "/c", f"docker context use {linux_context}"],
                check=True,
                capture_output=True,
            )

            # Verify the switch worked
            verify = subprocess.run(
                ["docker", "info"], capture_output=True, text=True, check=True
            )
            if "linux" in verify.stdout.lower():
                print(
                    f"Successfully switched to Linux containers using '{linux_context}' context"
                )
                return True

            warnings.warn(
                f"Failed to switch to Linux containers with context '{linux_context}': {verify.stdout}"
            )
            return False

        except subprocess.CalledProcessError as e:
            print(f"Error occurred: {e}")
            return False

        except subprocess.CalledProcessError as e:
            print(f"Failed to switch to Linux containers: {e}")
            if e.stdout:
                print(f"stdout: {e.stdout}")
            if e.stderr:
                print(f"stderr: {e.stderr}")
            return False
        except Exception as e:
            print(f"Unexpected error switching to Linux containers: {e}")
            return False

    @staticmethod
    def is_running() -> bool:
        """Check if Docker is running by pinging the Docker daemon."""
        if not DockerManager.is_docker_installed():
            return False
        try:
            # self.client.ping()
            client = docker.from_env()
            client.ping()
            print("Docker is running.")
            return True
        except docker.errors.DockerException as e:
            print(f"Docker is not running: {str(e)}")
            return False
        except Exception as e:
            print(f"Error pinging Docker daemon: {str(e)}")
            return False

    def start(self) -> bool:
        """Attempt to start Docker Desktop (or the Docker daemon) automatically."""
        print("Attempting to start Docker...")

        try:
            if sys.platform == "win32":
                docker_path = _win32_docker_location()
                if not docker_path:
                    print("Docker Desktop not found.")
                    return False
                subprocess.run(["start", "", docker_path], shell=True)
            elif sys.platform == "darwin":
                subprocess.run(["open", "/Applications/Docker.app"])
            elif sys.platform.startswith("linux"):
                subprocess.run(["sudo", "systemctl", "start", "docker"])
            else:
                print("Unknown platform. Cannot auto-launch Docker.")
                return False

            # Wait for Docker to start up with increasing delays
            print("Waiting for Docker Desktop to start...")
            attempts = 0
            max_attempts = 20  # Increased max wait time
            while attempts < max_attempts:
                attempts += 1
                if self.is_running():
                    print("Docker started successfully.")
                    return True

                # Gradually increase wait time between checks
                wait_time = min(5, 1 + attempts * 0.5)
                print(
                    f"Docker not ready yet, waiting {wait_time:.1f}s... (attempt {attempts}/{max_attempts})"
                )
                time.sleep(wait_time)

            print("Failed to start Docker within the expected time.")
            print(
                "Please try starting Docker Desktop manually and run this command again."
            )
        except KeyboardInterrupt:
            print("Aborted by user.")
            raise
        except Exception as e:
            print(f"Error starting Docker: {str(e)}")
        return False

    def validate_or_download_image(
        self, image_name: str, tag: str = "latest", upgrade: bool = False
    ) -> None:
        """
        Validate if the image exists, and if not, download it.
        If upgrade is True, will pull the latest version even if image exists locally.
        """
        ok = DockerManager.ensure_linux_containers_for_windows()
        if not ok:
            warnings.warn(
                "Failed to ensure Linux containers on Windows. This build may fail."
            )
        print(f"Validating image {image_name}:{tag}...")
        remote_image_hash_from_local_image: str | None = None
        remote_image_hash: str | None = None

        with get_lock(f"{image_name}-{tag}"):
            try:
                local_image = self.client.images.get(f"{image_name}:{tag}")
                print(f"Image {image_name}:{tag} is already available.")

                if upgrade:
                    remote_image = self.client.images.get_registry_data(
                        f"{image_name}:{tag}"
                    )
                    remote_image_hash = remote_image.id

                    try:
                        remote_image_hash_from_local_image = DISK_CACHE.get(
                            local_image.id
                        )
                    except KeyboardInterrupt:
                        raise
                    except Exception:
                        remote_image_hash_from_local_image = None
                        stack = traceback.format_exc()
                        warnings.warn(
                            f"Error getting remote image hash from local image: {stack}"
                        )
                    if remote_image_hash_from_local_image == remote_image_hash:
                        print(f"Local image {image_name}:{tag} is up to date.")
                        return

                    # Quick check for latest version
                    with Spinner(f"Pulling newer version of {image_name}:{tag}..."):
                        # This needs to be swapped out using the the command line interface AI!
                        # _ = self.client.images.pull(image_name, tag=tag)
                        cmd_list = ["docker", "pull", f"{image_name}:{tag}"]
                        cmd_str = subprocess.list2cmdline(cmd_list)
                        print(f"Running command: {cmd_str}")
                        subprocess.run(cmd_list, check=True)
                    print(f"Updated to newer version of {image_name}:{tag}")
                    local_image_hash = self.client.images.get(f"{image_name}:{tag}").id
                    if remote_image_hash is not None:
                        DISK_CACHE.put(local_image_hash, remote_image_hash)

            except docker.errors.ImageNotFound:
                print(f"Image {image_name}:{tag} not found.")
                with Spinner("Loading "):
                    cmd_list = ["docker", "pull", f"{image_name}:{tag}"]
                    cmd_str = subprocess.list2cmdline(cmd_list)
                    print(f"Running command: {cmd_str}")
                    subprocess.run(cmd_list, check=True)
                try:
                    local_image = self.client.images.get(f"{image_name}:{tag}")
                    local_image_hash = local_image.id
                    print(f"Image {image_name}:{tag} downloaded successfully.")
                except docker.errors.ImageNotFound:
                    warnings.warn(f"Image {image_name}:{tag} not found after download.")

    def tag_image(self, image_name: str, old_tag: str, new_tag: str) -> None:
        """
        Tag an image with a new tag.
        """
        image: Image = self.client.images.get(f"{image_name}:{old_tag}")
        image.tag(image_name, new_tag)
        print(f"Image {image_name}:{old_tag} tagged as {new_tag}.")

    def _container_configs_match(
        self,
        container: Container,
        command: str | None,
        volumes: dict | None,
        ports: dict | None,
    ) -> bool:
        """Compare if existing container has matching configuration"""
        try:
            # Check if container is using the same image
            container_image_id = container.image.id
            container_image_tags = container.image.tags

            # Simplified image comparison - just compare the IDs directly
            if not container_image_tags:
                print(f"Container using untagged image with ID: {container_image_id}")
            else:
                current_image = self.client.images.get(container_image_tags[0])
                if container_image_id != current_image.id:
                    print(
                        f"Container using different image version. Container: {container_image_id}, Current: {current_image.id}"
                    )
                    return False

            # Check command if specified
            if command and container.attrs["Config"]["Cmd"] != command.split():
                print(
                    f"Command mismatch: {container.attrs['Config']['Cmd']} != {command}"
                )
                return False

            # Check volumes if specified
            if volumes:
                container_mounts = (
                    {
                        m["Source"]: {"bind": m["Destination"], "mode": m["Mode"]}
                        for m in container.attrs["Mounts"]
                    }
                    if container.attrs.get("Mounts")
                    else {}
                )

                for host_dir, mount in volumes.items():
                    if host_dir not in container_mounts:
                        print(f"Volume {host_dir} not found in container mounts.")
                        return False
                    if container_mounts[host_dir] != mount:
                        print(
                            f"Volume {host_dir} has different mount options: {container_mounts[host_dir]} != {mount}"
                        )
                        return False

            # Check ports if specified
            if ports:
                container_ports = (
                    container.attrs["Config"]["ExposedPorts"]
                    if container.attrs["Config"].get("ExposedPorts")
                    else {}
                )
                container_port_bindings = (
                    container.attrs["HostConfig"]["PortBindings"]
                    if container.attrs["HostConfig"].get("PortBindings")
                    else {}
                )

                for container_port, host_port in ports.items():
                    port_key = f"{container_port}/tcp"
                    if port_key not in container_ports:
                        print(f"Container port {port_key} not found.")
                        return False
                    if not container_port_bindings.get(port_key, [{"HostPort": None}])[
                        0
                    ]["HostPort"] == str(host_port):
                        print(f"Port {host_port} is not bound to {port_key}.")
                        return False
        except KeyboardInterrupt:
            raise
        except docker.errors.NotFound:
            print("Container not found.")
            return False
        except Exception as e:
            stack = traceback.format_exc()
            warnings.warn(f"Error checking container config: {e}\n{stack}")
            return False
        return True

    def run_container_detached(
        self,
        image_name: str,
        tag: str,
        container_name: str,
        command: str | None = None,
        volumes: dict[str, dict[str, str]] | None = None,
        ports: dict[int, int] | None = None,
        remove_previous: bool = False,
    ) -> Container:
        """
        Run a container from an image. If it already exists with matching config, start it.
        If it exists with different config, remove and recreate it.

        Args:
            volumes: Dict mapping host paths to dicts with 'bind' and 'mode' keys
                    Example: {'/host/path': {'bind': '/container/path', 'mode': 'rw'}}
            ports: Dict mapping host ports to container ports
                    Example: {8080: 80} maps host port 8080 to container port 80
        """
        image_name = f"{image_name}:{tag}"
        try:
            container: Container = self.client.containers.get(container_name)

            if remove_previous:
                print(f"Removing existing container {container_name}...")
                container.remove(force=True)
                raise docker.errors.NotFound("Container removed due to remove_previous")
            # Check if configuration matches
            elif not self._container_configs_match(container, command, volumes, ports):
                print(
                    f"Container {container_name} exists but with different configuration. Removing and recreating..."
                )
                container.remove(force=True)
                raise docker.errors.NotFound("Container removed due to config mismatch")
            print(f"Container {container_name} found with matching configuration.")

            # Existing container with matching config - handle various states
            if container.status == "running":
                print(f"Container {container_name} is already running.")
            elif container.status == "exited":
                print(f"Starting existing container {container_name}.")
                container.start()
            elif container.status == "restarting":
                print(f"Waiting for container {container_name} to restart...")
                timeout = 10
                container.wait(timeout=10)
                if container.status == "running":
                    print(f"Container {container_name} has restarted.")
                else:
                    print(
                        f"Container {container_name} did not restart within {timeout} seconds."
                    )
                    container.stop(timeout=0)
                    print(f"Container {container_name} has been stopped.")
                    container.start()
            elif container.status == "paused":
                print(f"Resuming existing container {container_name}.")
                container.unpause()
            else:
                print(f"Unknown container status: {container.status}")
                print(f"Starting existing container {container_name}.")
                self.first_run = True
                container.start()
        except docker.errors.NotFound:
            print(f"Creating and starting {container_name}")
            out_msg = f"# Running in container: {command}"
            msg_len = len(out_msg)
            print("\n" + "#" * msg_len)
            print(out_msg)
            print("#" * msg_len + "\n")
            container = self.client.containers.run(
                image_name,
                command,
                name=container_name,
                detach=True,
                tty=True,
                volumes=volumes,
                ports=ports,
                remove=True,
            )
        return container

    def run_container_interactive(
        self,
        image_name: str,
        tag: str,
        container_name: str,
        command: str | None = None,
        volumes: dict[str, dict[str, str]] | None = None,
        ports: dict[int, int] | None = None,
    ) -> None:
        # Remove existing container
        try:
            container: Container = self.client.containers.get(container_name)
            container.remove(force=True)
        except docker.errors.NotFound:
            pass
        start_time = time.time()
        try:
            docker_command: list[str] = [
                "docker",
                "run",
                "-it",
                "--rm",
                "--name",
                container_name,
            ]
            if volumes:
                for host_dir, mount in volumes.items():
                    docker_command.extend(["-v", f"{host_dir}:{mount['bind']}"])
            if ports:
                for host_port, container_port in ports.items():
                    docker_command.extend(["-p", f"{host_port}:{container_port}"])
            docker_command.append(f"{image_name}:{tag}")
            if command:
                docker_command.append(command)
            cmd_str: str = subprocess.list2cmdline(docker_command)
            print(f"Running command: {cmd_str}")
            subprocess.run(docker_command, check=False)
        except subprocess.CalledProcessError as e:
            print(f"Error running Docker command: {e}")
            diff = time.time() - start_time
            if diff < 5:
                raise
            sys.exit(1)  # Probably a user exit.

    def attach_and_run(self, container: Container | str) -> RunningContainer:
        """
        Attach to a running container and monitor its logs in a background thread.
        Returns a RunningContainer object that can be used to stop monitoring.
        """
        if isinstance(container, str):
            container = self.get_container(container)

        assert container is not None, "Container not found."

        print(f"Attaching to container {container.name}...")

        first_run = self.first_run
        self.first_run = False

        return RunningContainer(container, first_run)

    def suspend_container(self, container: Container | str) -> None:
        """
        Suspend (pause) the container.
        """
        if isinstance(container, str):
            container_name = container
            container = self.get_container(container)
            if not container:
                print(f"Could not put container {container_name} to sleep.")
                return
        try:
            if platform.system() == "Windows":
                container.pause()
            else:
                container.stop()
                container.remove()
            print(f"Container {container.name} has been suspended.")
        except KeyboardInterrupt:
            print(f"Container {container.name} interrupted by keyboard interrupt.")
        except Exception as e:
            print(f"Failed to suspend container {container.name}: {e}")

    def resume_container(self, container: Container | str) -> None:
        """
        Resume (unpause) the container.
        """
        if isinstance(container, str):
            container = self.get_container(container)
        if not container:
            print(f"Could not resume container {container}.")
            return
        try:
            container.unpause()
            print(f"Container {container.name} has been resumed.")
        except Exception as e:
            print(f"Failed to resume container {container.name}: {e}")

    def get_container(self, container_name: str) -> Container | None:
        """
        Get a container by name.
        """
        try:
            return self.client.containers.get(container_name)
        except docker.errors.NotFound:
            return None

    def is_container_running(self, container_name: str) -> bool:
        """
        Check if a container is running.
        """
        try:
            container = self.client.containers.get(container_name)
            return container.status == "running"
        except docker.errors.NotFound:
            print(f"Container {container_name} not found.")
            return False

    def build_image(
        self,
        image_name: str,
        tag: str,
        dockerfile_path: Path,
        build_context: Path,
        build_args: dict[str, str] | None = None,
        platform_tag: str = "",
    ) -> None:
        """
        Build a Docker image from a Dockerfile.

        Args:
            image_name: Name for the image
            tag: Tag for the image
            dockerfile_path: Path to the Dockerfile
            build_context: Path to the build context directory
            build_args: Optional dictionary of build arguments
            platform_tag: Optional platform tag (e.g. "-arm64")
        """
        if not dockerfile_path.exists():
            raise FileNotFoundError(f"Dockerfile not found at {dockerfile_path}")

        if not build_context.exists():
            raise FileNotFoundError(
                f"Build context directory not found at {build_context}"
            )

        print(f"Building Docker image {image_name}:{tag} from {dockerfile_path}")

        # Prepare build arguments
        buildargs = build_args or {}
        if platform_tag:
            buildargs["PLATFORM_TAG"] = platform_tag

        try:
            cmd_list = [
                "docker",
                "build",
                "-t",
                f"{image_name}:{tag}",
            ]

            # Add build args
            for arg_name, arg_value in buildargs.items():
                cmd_list.extend(["--build-arg", f"{arg_name}={arg_value}"])

            # Add dockerfile and context paths
            cmd_list.extend(["-f", str(dockerfile_path), str(build_context)])

            cmd_str = subprocess.list2cmdline(cmd_list)
            print(f"Running command: {cmd_str}")

            # Run the build command
            # cp = subprocess.run(cmd_list, check=True, capture_output=True)
            proc: subprocess.Popen = subprocess.Popen(
                cmd_list, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            stdout = proc.stdout
            assert stdout is not None, "stdout is None"
            for line in iter(stdout.readline, b""):
                try:
                    line_str = line.decode("utf-8")
                    print(line_str, end="")
                except UnicodeDecodeError:
                    print("Error decoding line")
            rtn = proc.wait()
            if rtn != 0:
                warnings.warn(
                    f"Error building Docker image, is docker running? {rtn}, stdout: {stdout}, stderr: {proc.stderr}"
                )
                raise subprocess.CalledProcessError(rtn, cmd_str)
            print(f"Successfully built image {image_name}:{tag}")

        except subprocess.CalledProcessError as e:
            print(f"Error building Docker image: {e}")
            raise

    def purge(self, image_name: str) -> None:
        """
        Remove all containers and images associated with the given image name.

        Args:
            image_name: The name of the image to purge (without tag)
        """
        print(f"Purging all containers and images for {image_name}...")

        # Remove all containers using this image
        try:
            containers = self.client.containers.list(all=True)
            for container in containers:
                if any(image_name in tag for tag in container.image.tags):
                    print(f"Removing container {container.name}")
                    container.remove(force=True)

        except Exception as e:
            print(f"Error removing containers: {e}")

        # Remove all images with this name
        try:
            self.client.images.prune(filters={"dangling": False})
            images = self.client.images.list()
            for image in images:
                if any(image_name in tag for tag in image.tags):
                    print(f"Removing image {image.tags}")
                    self.client.images.remove(image.id, force=True)
        except Exception as e:
            print(f"Error removing images: {e}")


def main():
    # Register SIGINT handler
    # signal.signal(signal.SIGINT, handle_sigint)

    docker_manager = DockerManager()

    # Parameters
    image_name = "python"
    tag = "3.10-slim"
    # new_tag = "my-python"
    container_name = "my-python-container"
    command = "python -m http.server"

    try:
        # Step 1: Validate or download the image
        docker_manager.validate_or_download_image(image_name, tag, upgrade=True)

        # Step 2: Tag the image
        # docker_manager.tag_image(image_name, tag, new_tag)

        # Step 3: Run the container
        container = docker_manager.run_container_detached(
            image_name, tag, container_name, command
        )

        # Step 4: Attach and monitor the container logs
        running_container = docker_manager.attach_and_run(container)

        # Wait for keyboard interrupt
        while True:
            time.sleep(0.1)

    except KeyboardInterrupt:
        print("\nStopping container...")
        running_container.stop()
        container = docker_manager.get_container(container_name)
        docker_manager.suspend_container(container)

    try:
        # Suspend and resume the container
        container = docker_manager.get_container(container_name)
        docker_manager.suspend_container(container)

        input("Press Enter to resume the container...")

        docker_manager.resume_container(container)
    except Exception as e:
        print(f"An error occurred: {e}")


if __name__ == "__main__":
    main()
